// AKG4e3: 实时大气散射渲染实战
// https://zhuanlan.zhihu.com/p/595576594

// 张亚坤: 体渲染探秘（三）天空大气渲染
// https://zhuanlan.zhihu.com/p/419165090

// 未名客: 【实战】从零实现一套完整单次大气散射
// https://zhuanlan.zhihu.com/p/237502022

// 感谢 Coldnight冷夜 大佬在实现过程中提供的帮助

float getHeigth(vec3 p){
    return length(p) - earth_r;
}

float getRho(float h, float H){
    return min(fastExp(-(h / H)), 1.0);
}

vec3 RayleighCoefficient(float h){
    return RayleighSigma * getRho(h, H_R);
}

vec3 MieCoefficient(float h){
    return MieSigma * getRho(h, H_M);
}

vec3 MieAbsorption(float h){
    return MieAbsorptionSigma * getRho(h, H_M);
}

vec3 OzoneAbsorption(float h){
    float rho = max(0, 1.0 - abs(h - ozoneCenter) / ozoneWidth);
    return OzoneAbsorptionSigma * rho;
}

float RayleiPhase(float cos_theta){
    return (3.0 / (16.0 * PI)) * (1.0 + cos_theta * cos_theta);
}

float MiePhase(float cos_theta){
    float g = MIE_G;
    float g2 = g * g;

    // float a = 3.0 / (8.0 * PI);
    // float b = (1.0 - g2) / (2.0 + g2);
    // float c = 1.0 + cos_theta*cos_theta;
    // float d = pow(1.0 + g2 - 2*g*cos_theta, 1.5);
    // return a * b * (c / d);

    return (1 - g2) / (4.0 * PI * pow((1 + g2 - 2 * g * cos_theta), 3.0 / 2.0));
}

vec3 Scattering(float h, vec3 lightDir, vec3 worldDir){
    float cos_theta = dot(lightDir, worldDir);

    vec3 rayleigh = RayleighCoefficient(h) * RayleiPhase(cos_theta);
    vec3 mie = MieCoefficient(h) * MiePhase(cos_theta);

    return rayleigh + mie;
}

vec3 Transmittance(vec3 p1, vec3 p2, int N_SAMPLE){
    vec3 dir = normalize(p2 - p1);
    float distance = length(p2 - p1);
    float ds = distance / float(N_SAMPLE);
    vec3 sum = vec3(0.0);
    vec3 p = p1 + (dir * ds) * 0.5;

    for(int i=0; i<N_SAMPLE; i++){
        float h = getHeigth(p);

        vec3 scattering = RayleighCoefficient(h) + MieCoefficient(h);
        vec3 absorption = OzoneAbsorption(h) + MieAbsorption(h);
        vec3 extinction = scattering + absorption;

        sum += extinction * ds;
        p += dir * ds;
    }

    return fastExp(-sum);
}

void UvToTransmittanceLutParams(float bottomRadius, float topRadius, vec2 uv, out float mu, out float r){
    float x_mu = uv.x;
    float x_r = uv.y;

    float H = sqrt(max(0.0f, topRadius * topRadius - bottomRadius * bottomRadius));
    float rho = H * x_r;
    r = sqrt(max(0.0f, rho * rho + bottomRadius * bottomRadius));

    float d_min = topRadius - r;
    float d_max = rho + H;
    float d = d_min + x_mu * (d_max - d_min);
    mu = d == 0.0f ? 1.0f : (H * H - rho * rho - d * d) / (2.0f * r * d);

    mu = clamp(mu, -1.0f, 1.0f);
}

#define T1_T 0.0
vec3 DrawTransmittanceLut(){
    float bottomRadius = earth_r - T1_T;
    float topRadius = earth_r + atmosphere_h;

    float cos_theta = 0.0;
    float r = 0.0;
    UvToTransmittanceLutParams(bottomRadius, topRadius, texcoord, cos_theta, r);

    float sin_theta = sqrt(1.0 - cos_theta * cos_theta);
    vec3 worldDir = vec3(sin_theta, cos_theta, 0);
    vec3 eyePos = vec3(0, r, 0);

    float dis = RaySphereIntersection(eyePos, worldDir, vec3(0.0), topRadius).y;
    vec3 hitPoint = eyePos + worldDir * dis;

    return Transmittance(eyePos, hitPoint, 100);
}

vec2 GetTransmittanceLutUv(float bottomRadius, float topRadius, float mu, float r){
    float H = sqrt(max(0.0f, topRadius * topRadius - bottomRadius * bottomRadius));
    float rho = sqrt(max(0.0f, r * r - bottomRadius * bottomRadius));

    float discriminant = r * r * (mu * mu - 1.0f) + topRadius * topRadius;
	float d = max(0.0f, (-r * mu + sqrt(discriminant)));

    float d_min = topRadius - r;
    float d_max = rho + H;

    if(d > d_max) return vec2(-1.0, -1.0);

    float x_mu = (d - d_min) / (d_max - d_min);
    float x_r = rho / H;

    return vec2(x_mu, x_r);
}

vec3 TransmittanceToAtmosphere(vec3 p, vec3 dir){
    float bottomRadius = earth_r - T1_T;
    float topRadius = earth_r + atmosphere_h;

    vec3 upVector = normalize(p);
    float cos_theta = dot(upVector, dir);
    float r = length(p);

    vec2 uv = GetTransmittanceLutUv(bottomRadius, topRadius, cos_theta, r);
    if(outScreen(uv)) return BLACK;

    uv = uv * T1_RES + T1_UV;
    uv *= invViewSize;

#ifdef GBF
    return textureORB(gaux2, uv).rgb;
#endif

    return textureORB(colortex1, uv).rgb;
}

vec3 IntegralMultiScattering(vec3 samplePoint, vec3 lightDir){
    vec3 RandomSphereSamples[64] = {
        vec3(-0.7838,-0.620933,0.00996137),
        vec3(0.106751,0.965982,0.235549),
        vec3(-0.215177,-0.687115,-0.693954),
        vec3(0.318002,0.0640084,-0.945927),
        vec3(0.357396,0.555673,0.750664),
        vec3(0.866397,-0.19756,0.458613),
        vec3(0.130216,0.232736,-0.963783),
        vec3(-0.00174431,0.376657,0.926351),
        vec3(0.663478,0.704806,-0.251089),
        vec3(0.0327851,0.110534,-0.993331),
        vec3(0.0561973,0.0234288,0.998145),
        vec3(0.0905264,-0.169771,0.981317),
        vec3(0.26694,0.95222,-0.148393),
        vec3(-0.812874,-0.559051,-0.163393),
        vec3(-0.323378,-0.25855,-0.910263),
        vec3(-0.1333,0.591356,-0.795317),
        vec3(0.480876,0.408711,0.775702),
        vec3(-0.332263,-0.533895,-0.777533),
        vec3(-0.0392473,-0.704457,-0.708661),
        vec3(0.427015,0.239811,0.871865),
        vec3(-0.416624,-0.563856,0.713085),
        vec3(0.12793,0.334479,-0.933679),
        vec3(-0.0343373,-0.160593,-0.986423),
        vec3(0.580614,0.0692947,0.811225),
        vec3(-0.459187,0.43944,0.772036),
        vec3(0.215474,-0.539436,-0.81399),
        vec3(-0.378969,-0.31988,-0.868366),
        vec3(-0.279978,-0.0109692,0.959944),
        vec3(0.692547,0.690058,0.210234),
        vec3(0.53227,-0.123044,-0.837585),
        vec3(-0.772313,-0.283334,-0.568555),
        vec3(-0.0311218,0.995988,-0.0838977),
        vec3(-0.366931,-0.276531,-0.888196),
        vec3(0.488778,0.367878,-0.791051),
        vec3(-0.885561,-0.453445,0.100842),
        vec3(0.71656,0.443635,0.538265),
        vec3(0.645383,-0.152576,-0.748466),
        vec3(-0.171259,0.91907,0.354939),
        vec3(-0.0031122,0.9457,0.325026),
        vec3(0.731503,0.623089,-0.276881),
        vec3(-0.91466,0.186904,0.358419),
        vec3(0.15595,0.828193,-0.538309),
        vec3(0.175396,0.584732,0.792038),
        vec3(-0.0838381,-0.943461,0.320707),
        vec3(0.305876,0.727604,0.614029),
        vec3(0.754642,-0.197903,-0.62558),
        vec3(0.217255,-0.0177771,-0.975953),
        vec3(0.140412,-0.844826,0.516287),
        vec3(-0.549042,0.574859,-0.606705),
        vec3(0.570057,0.17459,0.802841),
        vec3(-0.0330304,0.775077,0.631003),
        vec3(-0.938091,0.138937,0.317304),
        vec3(0.483197,-0.726405,-0.48873),
        vec3(0.485263,0.52926,0.695991),
        vec3(0.224189,0.742282,-0.631472),
        vec3(-0.322429,0.662214,-0.676396),
        vec3(0.625577,-0.12711,0.769738),
        vec3(-0.714032,-0.584461,-0.385439),
        vec3(-0.0652053,-0.892579,-0.446151),
        vec3(0.408421,-0.912487,0.0236566),
        vec3(0.0900381,0.319983,0.943135),
        vec3(-0.708553,0.483646,0.513847),
        vec3(0.803855,-0.0902273,0.587942),
        vec3(-0.0555802,-0.374602,-0.925519),
    };
    const int N_DIRECTION = 64;
    const int N_SAMPLE = 32;
    const float uniform_phase = 1.0 / (4.0 * PI);
    const float sphereSolidAngle = 4.0 * PI / float(N_DIRECTION);
    
    vec3 G_2 = vec3(0, 0, 0);
    vec3 f_ms = vec3(0, 0, 0);

    for(int i=0; i<N_DIRECTION; i++){
        vec3 worldDir = RandomSphereSamples[i];
        float dis = RaySphereIntersection(samplePoint, worldDir, vec3(0.0), earth_r + atmosphere_h).y;
        float d = RaySphereIntersection(samplePoint, worldDir, vec3(0.0), earth_r).x;
        if(d > 0) dis = min(dis, d);
        float ds = dis / float(N_SAMPLE);

        vec3 p = samplePoint + (worldDir * ds) * 0.5;
        vec3 opticalDepth = vec3(0, 0, 0);

        for(int j=0; j<N_SAMPLE; j++)
        {
            float h = getHeigth(p);
            vec3 sigma_s = RayleighCoefficient(h) + MieCoefficient(h);
            vec3 sigma_a = OzoneAbsorption(h) + MieAbsorption(h);
            vec3 sigma_t = sigma_s + sigma_a;
            opticalDepth += sigma_t * ds;

            vec3 t1 = TransmittanceToAtmosphere(p, lightDir);
            vec3 s  = Scattering(h, lightDir, worldDir);
            vec3 t2 = fastExp(-opticalDepth);
            
            G_2  += t1 * s * t2 * uniform_phase * ds * 1.0;  
            f_ms += t2 * sigma_s * uniform_phase * ds;

            p += worldDir * ds;
        }
    }

    G_2 *= sphereSolidAngle;
    f_ms *= sphereSolidAngle;
    return 30.0 * G_2 * (1.0 / (1.0 - f_ms));
}

#define MS_SCALE 1.0
vec3 DrawMultiScatteringLut(){
    float mu_s = texcoord.x * 2.0 - 1.0;
    float r = mix(1.0, texcoord.y, MS_SCALE) * atmosphere_h + earth_r;

    float cos_theta = mu_s;
    float sin_theta = sqrt(1.0 - cos_theta * cos_theta);
    vec3 lightDir = vec3(sin_theta, cos_theta, 0);
    vec3 p = vec3(0, r, 0);

    return IntegralMultiScattering(p, lightDir);
}

vec3 GetMultiScattering(float h, vec3 p, vec3 lightDir){
    float cosSunZenithAngle = dot(normalize(p), lightDir);
    vec2 uv = vec2(cosSunZenithAngle * 0.5 + 0.5, mix(1.0, h / atmosphere_h, 1.0 / MS_SCALE));
    if(outScreen(uv)) return BLACK;

    uv = uv * MS_RES + MS_UV;
    uv *= invViewSize;

#ifdef GBF
    return textureORB(gaux2, uv).rgb * 0.02;
#endif

    return textureORB(colortex1, uv).rgb * 0.02;
}

void getDensity(float h, out float dRayleigh_c, out float dMie_c){
    dRayleigh_c = getRho(h, H_R);
    dMie_c = getRho(h, H_M);
}

vec3 ExtinctionT2(float h, float dRayleigh_c, float dMie_c,
                inout float dRayleigh, inout float dMie, inout float dOzone){
    dRayleigh += dRayleigh_c;
    dMie += dMie_c;
    dOzone += max(0, 1.0 - (abs(h - ozoneCenter) / ozoneWidth));
    return dRayleigh * RayleighSigma
        + dMie * (MieSigma + MieAbsorptionSigma)
        + dOzone * OzoneAbsorptionSigma;
}

mat2x3 AtmosphericScattering(vec3 worldPos, vec3 lightDir, vec3 I, float mieAmount, const int N_SAMPLES, const int lutSampleGap){
    float dis = length(worldPos);
    vec3 worldDir = normalize(worldPos);

    float ds = dis / float(N_SAMPLES);
    vec3 p = earthPos /*+ worldDir * ds * temporalBayer64(texcoord * viewSize)*/;

    vec3 inScatteringR = BLACK;
    vec3 inScatteringM = BLACK;
    vec3 multiScatteringR = BLACK;
    vec3 multiScatteringM = BLACK;

    float dRayleigh = 0.0;
    float dMie = 0.0;
    float dRayleigh_c = 0.0;
    float dMie_c = 0.0;
    float dOzone = 0.0;

    float h = getHeigth(p);
    vec3 t1 = TransmittanceToAtmosphere(p, lightDir);
    vec3 G_ALL = GetMultiScattering(h, p, lightDir);

    for(int i = 0; i < N_SAMPLES; i++){
        h = getHeigth(p);

        if(i % lutSampleGap == 1){
            t1 = TransmittanceToAtmosphere(p, lightDir);
            G_ALL = GetMultiScattering(h, p, lightDir);
        }

        p += worldDir * ds;

        // if(t1.r + t1.g + t1.b + G_ALL.r + G_ALL.g + G_ALL.b < 0.0001) continue;
        
        getDensity(h, dRayleigh_c, dMie_c);
        vec3 opticalDepthT2 = ExtinctionT2(h, dRayleigh_c, dMie_c, dRayleigh, dMie, dOzone) * ds;
        vec3 t2 = fastExp(-opticalDepthT2);

        vec3 t = t1 * t2;

        inScatteringR += t * dRayleigh_c;
        inScatteringM += t * dMie_c;

        G_ALL *= t2;
        multiScatteringR += G_ALL * dRayleigh_c;
        multiScatteringM += G_ALL * dMie_c;    
    }

    float cos_theta = dot(worldDir, lightDir);
    vec3 inScattering = inScatteringR * RayleighSigma * RayleiPhase(cos_theta) * 1.0 
                        + inScatteringM * MieSigma * MiePhase(cos_theta) * 1.0;
    vec3 multiScattering = multiScatteringR * RayleighSigma * 1.0
                    + multiScatteringM * MieSigma * mieAmount * MIE_STRENGTHNESS;
	
    return mat2x3(inScattering * ds * IncomingLight, multiScattering * ds * I);
}



void Extinction(float h,
                inout float dRayleigh, inout float dMie, inout float dOzone){
    dRayleigh += getRho(h, H_R);
    dMie += getRho(h, H_M);
    dOzone += max(0, 1.0 - abs(h - ozoneCenter) / ozoneWidth);
}

vec3 Transmittance1(vec3 startPos, vec3 endPos, float stepCount){
    vec3 direction = normalize(endPos - startPos);
    float dist = length(endPos - startPos);
    float stepSize = dist / stepCount;
    vec3 stepVec = direction * stepSize;

    vec3 p = startPos + 0.5 * stepVec;

    float dRayleigh = 0.0;
    float dMie = 0.0;
    float dOzone = 0.0;

    for(int i = 0; i < stepCount; i++){
        float h = getHeigth(p);
        Extinction(h, dRayleigh, dMie, dOzone);
        p += stepVec;
    }
    vec3 opticalDepth = dRayleigh * RayleighSigma
                + dMie * (MieSigma + MieAbsorptionSigma)
                + dOzone * OzoneAbsorptionSigma;
    vec3 t = fastExp(-opticalDepth * stepSize);

    return t;
}



#define SUN_COLOR_UV ivec2(1, viewSize.y - 1)
#define SKY_COLOR_UV ivec2(3, viewSize.y - 1)

vec3 getSunColor(){
    vec3 sunColor = texelFetch(colortex1, SUN_COLOR_UV, 0).rgb;
    return sunColor;
}

vec3 getSkyColor(){
    vec3 skyColor = texelFetch(colortex1, SKY_COLOR_UV, 0).rgb;
    return skyColor;
}

#define ZENITH_COLOR_UV ivec2(5, viewSize.y - 1)
#define HORIZON_COLOR_UV ivec2(7, viewSize.y - 1)

vec3 getZenithColor(){
    vec3 zenithColor = texelFetch(colortex1, ZENITH_COLOR_UV, 0).rgb;
    return zenithColor;
}

vec3 getHorizonColor(){
    vec3 horizonColor = texelFetch(colortex1, HORIZON_COLOR_UV, 0).rgb;
    return horizonColor;
}

/*
vec3 AtmosphericScattering(vec3 worldDir, vec3 lightDir){
    const int N_SAMPLE = 64;

    float d_p2a = RaySphereIntersection(earthPos, worldDir, vec3(0.0), earth_r + atmosphere_h).y;
    // float d_p2e = RaySphereIntersection(earthPos, worldDir, vec3(0.0), earth_r).x;
    float dis = d_p2a;
    // if(d_p2a < 0.0) return BLACK; 
    // if(d_p2e > 0.0) dis = min(dis, d_p2e);

    float ds = dis / float(N_SAMPLE);
    vec3 p = earthPos + worldDir * ds * 0.0;

    vec3 color = BLACK;
    vec3 opticalDepth = BLACK;

    for(int i = 0; i < N_SAMPLE; i++){
        float h = getHeigth(p);

        vec3 t1 = TransmittanceToAtmosphere(p, lightDir);
        
        vec3 s = Scattering(h, worldDir, lightDir);
        
        vec3 extinction = RayleighCoefficient(h) + MieCoefficient(h) +
                        OzoneAbsorption(h) + MieAbsorption(h);
        opticalDepth += extinction * ds;
        vec3 t2 = fastExp(-opticalDepth);

        vec3 inScattering = t1 * s * t2 * ds;
        color += inScattering;

        vec3 multiScattering = GetMultiScattering(h, p, lightDir) 
                                * (RayleighCoefficient(h) + MieCoefficient(h));
        color += multiScattering * t2 * ds;

        p += worldDir * ds;
    }

	return color * vec3(1.0) * 8.0;
}
*/